Option Explicit On 
Imports System.Threading
'---------------------------------------------------------------
 ' Module Name: "SoftTimer"
 '
 ' This creates and manages software timer objects
 '
 ' References: See RR model XXYYZZ.MDL
 '
 ' Created By:
 '
 ' $Workfile: $
 ' $Revision: $
 ' $Date: $
 ' $Log: $
 '---------------------------------------------------------------
Public Class CSoftTimer

    Private lStart As Long          ' System Time at "Clock Start"
    Private lElapsed As Long        ' Cumulative Time Elapsed
    Private ModuleName As String
    Private bTime As Boolean        ' Indicates Timing in Progress
    Public TimeOutValue As Integer  ' Used for time-out checking

    Public Sub New(ByVal iTimeout As Integer)
        TimeOutValue = iTimeout
        Me.ClockStart()
    End Sub

    Public Sub New(ByVal bStart As Boolean)
        If bStart Then Me.ClockStart()
    End Sub

    ' This sets a "one shot" timer, which disables
    ' itself at the end of the specified interval,
    ' which is specified in seconds.
    Public Sub Wait(ByVal sInterval As Single)
        Dim lTarget As Long

        If sInterval < 0.01 Then Exit Sub   ' Ignore delay this short
        lTarget = Environment.TickCount + (1000 * sInterval)
        ' Calculate end time
        While Environment.TickCount < lTarget
            Thread.Sleep(200)               ' Wait & Yeild while waiting
            Thread.Yield()
        End While
    End Sub

    Public ReadOnly Property Elapsed() As Long
        Get
            ' Return the Current elapsed time in ms.
            ' If running, include the current time since start,
            ' else just the elapsed time when stopped
            If bTime Then
                Elapsed = lElapsed + Environment.TickCount - lStart
            Else
                Elapsed = lElapsed
            End If

        End Get
    End Property
    ' Reset and Stop the Clock
    Public Sub Reset()

        lElapsed = 0            ' Set Elapsed Time to Zero
        bTime = False           ' Not Running
    End Sub

    ' Start the Clock (If Not Already Running)
    ' Optionally reset the elapsed time counter at the same time
    Public Sub ClockStart(Optional ByVal bReset As Object = False)

        If bReset Then                      ' Reset the elapsed time
            lElapsed = 0
        End If
        If (Not bTime) Then                 ' Not yet running
            lStart = Environment.TickCount
            bTime = True                    ' Indicate timing in progress
        End If
    End Sub
    ' Stop the Clock and add cumulative time to elapsed time counter
    Public Sub ClockStop()

        lElapsed = lElapsed + Environment.TickCount - lStart
        ' Set Elapsed Time
        bTime = False           ' Not Running
    End Sub
    Public Sub New()
        ModuleName = TypeName(Me)
        Me.ClockStart()
    End Sub

    Public ReadOnly Property Timeout() As Boolean
        Get
            Return (Me.Elapsed > TimeOutValue)
        End Get
    End Property
End Class
