﻿Imports System.Drawing
Imports System.Reflection
Imports System.Reflection.MethodBase
Imports System.Runtime.CompilerServices
Imports System.Windows.Forms
Imports Microsoft.Office.Tools
Imports ErrorLog
Imports ErrorLog.LogStates
Imports Newtonsoft.Json
Imports Newtonsoft.Json.Linq
Imports System.Reflection.PropertyInfo
Imports System.Collections.Specialized
Imports System.Text.RegularExpressions
Imports System.Globalization
Imports System.Globalization.CultureInfo

Public Module modExtensions

    Private sModule As String = "modExtensions"
    Public Event ErrorRaised(oDetail As ErrorDetail)

#Region "Rectangle Functions"
    <Extension()>
    Public Function Bottom(ByVal oRect As Rectangle) As Double
        ' Calculates the "bottom" position of the rectangle
        Return oRect.Top + oRect.Height
    End Function

    <Extension()>
    Public Function Right(ByVal oRect As Rectangle) As Double
        ' Calculates the "right" position of the rectangle
        Return oRect.Left + oRect.Width
    End Function
#End Region

#Region "String Functions"
    <Extension()>
    Function Remove(sXML As String, sStart As String, sEnd As String, bRemoveEnd As Boolean) As String
        ' Removes a specified section of text from an HTML/XML file
        Try
            Dim iPos As Integer, iPos2 As Integer
            Do
                iPos = sXML.IndexOf(sStart)                                ' Reduce any amazon URLs to ""
                If iPos > 0 Then
                    iPos2 = sXML.IndexOf(sEnd, iPos)
                    If bRemoveEnd Then
                        sXML = sXML.Substring(0, iPos) & sXML.Substring(iPos2 + sEnd.Length)
                    Else
                        sXML = sXML.Substring(0, iPos) & sXML.Substring(iPos2)
                    End If
                End If
            Loop Until iPos <= 0
            Return sXML
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function Truncate(ByVal aString As String, maxLen As Integer) As String
        ' Extend the string type with a working truncate function
        If maxLen > aString.Length Then
            Return aString
        Else
            Return aString.Substring(0, maxLen)
        End If
    End Function

    <Extension()>
    Public Function ExtractSegment(sContent As String, iStart As Integer, sStart As String, sEnd As String,
                                   Optional bIncludeDelimiters As Boolean = False) As String
        ' Extract trimmed text delimited by two strings after a given start point
        ' Returns blank if delimiters cannot be found
        Try
            Dim iPos, iPos2, iPos3 As Integer
            Dim sSegment As String = ""

            iPos = InStr(iStart, sContent, sStart)
            If iPos > 0 Then
                iPos2 = iPos + Len(sStart) - 1
                iPos3 = InStr(iPos2, sContent, sEnd)
                If iPos3 > 0 Then
                    If bIncludeDelimiters Then
                        sSegment = sContent.Substring(iPos - 1, iPos3 - iPos + Len(sEnd))
                    Else
                        sSegment = sContent.Substring(iPos2, iPos3 - iPos2 - 1)
                    End If
                    sSegment = sSegment.Trim
                End If
            End If
            Return sSegment
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function Strpos(sContent As String, sTarget As String, Optional iStart As Integer = 1) As String
        ' Find position of text in string
        Try
            Return InStr(iStart, sContent, sTarget)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Function Left(sTarget As String, iLength As Integer) As String
        ' Return Left-most iLength characters, or whole string if shorter
        Try
            If sTarget.Length < iLength Then
                Return sTarget
            Else
                Return sTarget.Substring(0, iLength)
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Function Right(sTarget As String, iLength As Integer) As String
        ' Return Right-most iLength characters, or whole string if shorter
        Try
            If sTarget.Length < iLength Then
                Return sTarget
            Else
                Return sTarget.Substring(sTarget.Length - iLength)
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function


    <Extension()>
    Public Function ReplaceAt(sContent As String, iStart As Integer, iLength As Integer, sNewText As String) As String
        ' Replace text at a nominated position
        Try
            Return sContent.Remove(iStart, Math.Min(iLength, sContent.Length - iStart)).Insert(iStart, sNewText)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function Capitalise(sContent As String) As String
        ' Capitalise
        Try
            If sContent = "" Or sContent Is Nothing Then
                Return ""
            Else
                Return sContent.Substring(0, 1).ToUpper & sContent.Substring(1).ToLower
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function ReplaceInArray(ByRef oArray() As String, sTarget As String, sReplacement As String,
                                   Optional bCaseInsensitive As Boolean = False)
        Try
            For i = oArray.GetLowerBound(0) To oArray.GetUpperBound(0)
                If bCaseInsensitive Then
                    oArray(i) = oArray(i).ToString.ReplaceCaseInsensitive(sTarget, sReplacement)
                Else
                    oArray(i) = oArray(i).ToString.Replace(sTarget, sReplacement)
                End If
            Next
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function ReplaceCaseInsensitive(sSubject As String, sTarget As String, sReplacement As String) As String
        Try
            Return Regex.Replace(sSubject, Regex.Escape(sTarget), Regex.Replace(sReplacement, "\\$[0-9]+", "$$$0"), RegexOptions.IgnoreCase)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    Function GetBlankString(oObject As Object) As String
        ' Simple function to evaluate if a DB column is neither null nor empty, and then return an empty string if it is, else the value
        Try
            If IsDBNull(oObject) Or oObject Is Nothing Then
                Return ""
            Else
                Return oObject.ToString
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Function WildCardMatch(sTarget As String, sPattern As String) As Boolean
        ' Assess if target matches pattern, allowing for * wildcards
        Try
            Dim sParts() As String = sPattern.Split("*")
            If sParts.Count = 1 Then                        ' No wildcard - must be complete match
                If sTarget = sParts(0) Then Return True
            ElseIf sParts.Count = 2 Then                    ' We have a beginning and end to match
                If sTarget.StartsWith(sParts(0)) And sTarget.EndsWith(sParts(1)) Then Return True
            ElseIf sParts.Count = 3 Then                    ' We have a beginning and end to match
                If sTarget.StartsWith(sParts(0)) And sTarget.Contains(sParts(1)) And sTarget.EndsWith(sParts(2)) Then Return True
            Else
                Throw New Exception(GetCurrentMethod().Name & ": Unable to support this combination of wildcards: " & sPattern)
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Function IsValidNumberFormat(sFormat As String) As Boolean
        ' Test if string is a valid numeric/date formatting string
        Try
            Dim iValue As Integer = 100
            Dim dValue As Date = Now
            Dim bValue As Boolean = IsNumeric(Microsoft.VisualBasic.Format(iValue, sFormat))
            Dim bValue2 As String = IsDate(Microsoft.VisualBasic.Format(dValue, sFormat))
            If bValue Or bValue2 Then       ' Invalid Format String
                Return False
            Else
                Return True
            End If
        Catch ex As Exception               ' Ignore exception, just return false
            Return False
        End Try
    End Function

#End Region

#Region "Boolean Functions"
    Public Function IsBoolean(sValue As String) As Boolean
        ' Returns true if value can parse as boolean
        Try
            Dim bValue As Boolean
            Return Boolean.TryParse(sValue, bValue)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function ToIBoolExtended(sValue As String, Optional bIdentity As Boolean = False) As Integer
        ' Returns Integer Boolean equivalent of string, or an identity value if blank
        Try
            If sValue = "" Then
                Return If(bIdentity, 1, 0)
            ElseIf {"N", "NO", "F", "FALSE", "0", "OFF"}.Contains(sValue.ToUpper) Then
                Return 0
            ElseIf {"Y", "YES", "T", "TRUE", "1", "ON"}.Contains(sValue.ToUpper) Then
                Return 1
            Else
                Throw New Exception("Invalid boolean expression: " & sValue)
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function
#End Region

#Region "Date Functions"
#End Region

#Region "Object Functions"
    <Extension()>
    Function NotNullorEmpty(oObject As Object) As Boolean
        ' Simple function to evaluate if a DB column is neither null nor empty
        Try
            If Not IsDBNull(oObject) And Not (oObject Is Nothing) Then
                If oObject <> "" Then Return True
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Function GetBlankValue(oObject As Object) As Object
        ' Simple function to evaluate if a DB column is neither null nor empty, and then return an empty string if it is, else the value
        Try
            If IsDBNull(oObject) Or oObject Is Nothing Then
                Return ""
            Else
                Return oObject
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    <Extension()>
    Public Function NullToString(Of T)(this As T) As String
        If this Is Nothing Then
            Return String.Empty
        End If
        Return this.ToString()
    End Function

    <Extension()>
    Function CloneObject(oOriginal As Object) As Object
        ' Return a clone of oOriginal with same properties
        Try
            Dim oClone As Object = Activator.CreateInstance(oOriginal.GetType())
            For Each p As System.Reflection.PropertyInfo In oOriginal.GetType().GetProperties()
                If p.CanRead Then
                    oClone.GetType().GetProperty(p.Name).SetValue(oClone, p.GetValue(oOriginal, Nothing), Nothing)
                End If
            Next
            Return oClone
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Function

    '<Extension()>
    'Public Function GetProperty(oObject As Object, sName As String) As Object
    '    Try
    '        Return oObject.GetType().GetProperty(sName).GetValue(oObject)
    '    Catch ex As Exception
    '        Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message & " - " & sName, ex)
    '    End Try
    'End Function

#End Region

#Region "Control Save/Restore"
    <Extension()>
    Sub SaveText(oControl As Control)
        ' Save control text to application settings
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.Text)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SavePosition(oControl As Control)
        ' Save control position to application settings
        Try
            If oControl.Left >= 0 And oControl.Top >= 0 Then
                SaveSetting(Application.ProductName, "Settings", oControl.Name & "Left", oControl.Left)
                SaveSetting(Application.ProductName, "Settings", oControl.Name & "Top", oControl.Top)
            End If
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SavePosition(oControl As Splitter)
        ' Save control position to application settings
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name & "Pos", oControl.SplitPosition)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveSize(oControl As Control)
        ' Save control size to application settings
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name & "Width", oControl.Width)
            SaveSetting(Application.ProductName, "Settings", oControl.Name & "Height", oControl.Height)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveWindowState(oControl As Form)
        ' Save Form window state
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name & "WindowState", oControl.WindowState)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveValue(oControl As NumericUpDown)
        ' Save control value to settings
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.Value)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveValue(oControl As DateTimePicker)
        ' Save control value to settings
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.Value)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveValue(oControl As CheckBox)
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.Checked)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveValue(oControl As RadioButton)
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.Checked)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub SaveValue(oControl As ComboBox)
        Try
            SaveSetting(Application.ProductName, "Settings", oControl.Name, oControl.SelectedItem)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreText(oControl As Control)
        Try
            oControl.Text = GetSetting(Application.ProductName, "Settings", oControl.Name)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestorePosition(oControl As Control)
        Try
            oControl.Left = GetSetting(Application.ProductName, "Settings", oControl.Name & "Left", oControl.Left)
            oControl.Top = GetSetting(Application.ProductName, "Settings", oControl.Name & "Top", oControl.Top)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestorePosition(oControl As Splitter)
        Try
            oControl.SplitPosition = GetSetting(Application.ProductName, "Settings", oControl.Name & "Pos", oControl.SplitPosition)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreSize(oControl As Control)
        Try
            oControl.Width = GetSetting(Application.ProductName, "Settings", oControl.Name & "Width", oControl.Width)
            oControl.Height = GetSetting(Application.ProductName, "Settings", oControl.Name & "Height", oControl.Height)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreWindowState(oControl As Form)
        Try
            oControl.WindowState = GetSetting(Application.ProductName, "Settings", oControl.Name & "WindowState", FormWindowState.Normal)
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub


    <Extension()>
    Sub RestoreValue(oControl As NumericUpDown)
        Try
            Dim sValue As String = GetSetting(Application.ProductName, "Settings", oControl.Name)
            If IsNumeric(sValue) Then oControl.Value = sValue
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreValue(oControl As DateTimePicker)
        Try
            Dim sValue As String = GetSetting(Application.ProductName, "Settings", oControl.Name)
            If IsDate(sValue) Then oControl.Value = sValue
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreValue(oControl As CheckBox)
        Try
            Dim sValue As String = GetSetting(Application.ProductName, "Settings", oControl.Name)
            If sValue <> "" Then oControl.Checked = sValue
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreValue(oControl As RadioButton)
        Try
            Dim sValue As String = GetSetting(Application.ProductName, "Settings", oControl.Name)
            If sValue <> "" Then oControl.Checked = sValue
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub

    <Extension()>
    Sub RestoreValue(oControl As ComboBox)
        Try
            Dim sValue As String = GetSetting(Application.ProductName, "Settings", oControl.Name)
            If sValue <> "" Then oControl.SelectedItem = sValue
        Catch ex As Exception
            Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
        End Try
    End Sub
#End Region

#Region "Jaccard String Matching"
    <Extension()>
    Public Function Matches(ByVal aString As String, ByVal bString As String) As Boolean
        ' Performs a Jaccard match on the text content of two strings, and returns an assessment of whether it's a valid match
        Dim siMatchLevel As Single = JaccardIndex(Tokens(aString), Tokens(bString))
        If siMatchLevel > 0.8 Then
            Return True
        End If
        Console.WriteLine(aString)
    End Function

    Function JaccardIndex(oList1 As List(Of String), oList2 As List(Of String)) As Single
        ' Calculate modified Jaccard similarity coefficient for two lists of string tokens
        Try
            Dim m1 As Integer, m2 As Integer
            Dim siIndex As Single

            For Each s As String In oList1                  ' Count matches of list 1 in list 2
                If oList2.Contains(s) Then m1 += 1
            Next
            For Each s As String In oList2                  ' Count matchesof list 2 in list 1
                If oList1.Contains(s) Then m2 += 1
            Next

            siIndex = ((m1 + m2 - Math.Abs(oList1.Count - oList2.Count)) / (oList1.Count + oList2.Count))
            If siIndex < 0 Then siIndex = 0
            Return siIndex
        Catch ex As Exception
            MsgBox(ex.Message)
        End Try
    End Function

    Function Tokens(sDesc As String) As List(Of String)
        ' Split input string into list of non-empty tokens
        Try
            Dim sSeparators As [Char]() = {" "c, "."c, ","c, ":"c, ";"c, """"c, "-"c, "/"c, "("c, ")"c, "["c, "]"c}
            Dim oList As New List(Of String)
            Dim split As String() = sDesc.ToLower.Split(sSeparators)
            For Each s As String In split
                s = s.Trim()                        ' Get rid of trailing whitespace
                If s = "" Then                      ' Blank, ignore
                    'ElseIf sNoise.Contains(s) Then      ' Noise word, just ignore
                    'ElseIf sAbbrevs.Contains(s) Then    ' Abbreviation - expand it
                    '    oList.Add(sFull(sAbbrevs.IndexOf(s)))
                Else                                ' Normal, just add
                    oList.Add(s.Trim())
                End If
            Next s
            Return oList
        Catch ex As Exception
            MsgBox(ex.Message)
        End Try
    End Function
#End Region

#Region "JSON Functions"
    <Extension()>
    Function GNN(oCollection As Collection, sKey As String) As Object
        If oCollection.Contains(sKey) Then
            Return oCollection(sKey)
        Else
            Return ""
        End If
    End Function

    <Extension()>
    Function GNN(oCollection As JToken, sKey As String) As Object
        Try
            If oCollection Is Nothing Then
                Return ""
            End If
            Dim oObject = oCollection(sKey)
            If Not oObject Is Nothing Then
                Return oObject
            Else
                Return ""
            End If
        Catch ex As Exception
            Debug.Print("GNN on JToken " & ex.Message)
            Return ""
        End Try
    End Function

    <Extension()>
    Function GNN(oCollection As JObject, sKey As String) As Object
        Try
            If oCollection.ContainsKey(sKey) Then
                Dim oObject As JToken = oCollection(sKey)
                If oObject.HasValues Then
                    Return oObject
                End If
                Dim oValue = oObject.Value(Of String)
                If Not oValue Is Nothing Then
                    Return oObject
                Else
                    Return ""
                End If
            Else
                Return ""
            End If
        Catch ex As Exception
            Debug.Print("GNN on JObject " & ex.Message)
        End Try
    End Function

    <Extension()>
    Function GNN(oCollection As NameValueCollection, sKey As String) As Object
        If oCollection.AllKeys.Contains(sKey) Then
            Return oCollection(sKey)
        Else
            Return ""
        End If
    End Function


    <Extension()>
    Function GetByPath(oJA As JObject, sPath As String) As String
        ' Extract a value from the JObject using the specified .-separated path, handling missing & blank tokens
        Try
            Dim oToken As JToken = oJA
            For Each sStep As String In sPath.Split(".")    ' Drill down through path
                oToken = oToken(sStep)                      ' Stop if token goes to nothing
            Next
            If Not oToken Is Nothing Then                   ' We found a valid destination element
                Return oToken.ToString
            Else
                Return ""
            End If
        Catch ex As Exception
            If ex.Message.StartsWith("Cannot access child value") Then
                Return "Missing"                            ' Return identifiable status for missing value
            Else
                Throw New Exception(sModule & "." & GetCurrentMethod().Name & ":" & ex.Message, ex)
            End If
        End Try
    End Function

    <Extension()>
    Function GetKeys(oJA As JObject) As List(Of String)
        Dim keys As List(Of String) = oJA.Properties().Select(Function(p) p.Name).ToList()
        Return keys
    End Function
#End Region
End Module
